import logging
import pandas as pd

from datetime import datetime
from tagr.config import OBJECTS, RECOGNIZED_DTYPES
from tagr.storage.aws import Aws
from tagr.storage.azure import Azure
from tagr.storage.gcp import Gcp
from tagr.storage.local import Local

logger = logging.getLogger("tagging_artifact")


class Artifact:
    def __init__(self, val, obj_name, dtype=None):
        self.val = val
        self.obj_name = obj_name

        if dtype:
            self.dtype = dtype
            self.is_recognized_dtypes(self.dtype)
        else:
            self.dtype = OBJECTS[obj_name]

        self.check_expected_type()

    def __repr__(self):
        return "val: {0}, obj_name: {1}, dtype: {2}".format(
            self.val, self.obj_name, self.dtype
        )

    def check_expected_type(self):
        if self.dtype == "dataframe":
            self.evaluate_type(self.val, pd.DataFrame)
        elif self.dtype == "primitive":
            self.evaluate_type(self.val, (int, float, str, bool))

    @staticmethod
    def evaluate_type(val, expected_type):
        if not isinstance(val, expected_type):
            raise TypeError(
                "TypeError: artifact type did not match expected {} type".format(
                    str(expected_type)
                )
            )

    @staticmethod
    def is_recognized_dtypes(dtype):
        if not dtype in RECOGNIZED_DTYPES:
            raise ValueError("dtype must be of {}".format(", ".join(RECOGNIZED_DTYPES)))


class Tagr(object):
    def __init__(self):
        self.queue = {}
        self.cust_queue = {}
        self.storage_provider = Local()

        self.col_types_dict = {}
        self.col_stats_dict = {}

    def save(self, artifact, obj_name: str, dtype: str = None):
        """
        Tags a variable as something to be saved

        Parameters
        ----------
        artifact: variable to store
        obj: str. type of variable to be store.
            Choice of config.OBJECTS
            If you pass a string not in config.OBJECTS
            Tagr will store the object under the cust dir
        dtype: type or None, Default None
            used for storing custom variables not in OBJECTS
            Behavior as follows:

            * If `dtype` = 'string': saved to metadata json file as String
            * If `dtype` = 'df': saved as parquet to metadata store provider
            * If `dtype` = 'int' or 'float' or 'str': saved to metadata
             json file as int or float or str respectively. Need to retrieve
                entire json file to access any values
            * If `dtype` = 'metric': saved as readily accessible pickle
                value via metadata provider client
            * If `dtype`= 'viz' or 'other: saved as pickle

        Returns
        -------
        `artifact` to preserve declarative interface
        """
        if obj_name in OBJECTS:
            save_obj = Artifact(artifact, obj_name)
            self.queue[obj_name] = save_obj
        elif not dtype:
            raise ValueError("dtype must be provided if custom obj")
        else:
            save_obj = Artifact(artifact, obj_name, dtype)
            self.queue[obj_name] = save_obj

        return artifact

    def rm(self, obj_name):
        """
        Remove the supplied object from the dictionary of saved objects

        Parameter
        ---------
        obj_name: str. Name of object to be removed

        """
        if obj_name in self.queue:
            del self.queue[obj_name]
        else:
            raise KeyError("{} is not a key in the queue".format(obj_name))

    def ret_queue(self) -> dict:
        return self.queue

    def summary(self):
        """
        Returns pd.Dataframe of tagged variables and their values
        """
        data = []
        for k, artifact in self.queue.items():
            data.append([artifact.obj_name, artifact.val, artifact.dtype])

        return pd.DataFrame(data, columns=["obj_name", "val", "dtype"])

    def flush(self, proj, experiment, tag=None, storage="local"):
        """
        Pushes all variables from `queue` to metadata store.
        Generates metadata for artifacts of type pd.Dataframe in JSON
        format

        Parameters
        ----------
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        storage: destination for experiment data to be dumped ('aws', 'gcp', 'azure', 'local')
            - for storage, asssume local by default if destination not provided
        """

        # use datetime as index if tag name not provided
        if not tag:
            logger.info("using datetime as tag")
            tag = str(datetime.utcnow())

        if storage == "aws":
            self.storage_provider = Aws()
        elif storage == "gcp":
            self.storage_provider = Gcp()
        elif storage == "azure":
            self.storage_provider = Azure()
        elif storage == "local":
            self.storage_provider = Local()
            # if folder directory doesnt exist, then create new directory
            self.storage_provider.build_path(proj, experiment, tag)

        experiment_params = {"proj": proj, "experiment": experiment, "tag": tag}

        #####################
        # generate metadata #
        #####################
        summary = self.summary()
        self._flush_dfs(summary, experiment_params)
        self._flush_metadata_json(summary, experiment_params)
        self._flush_non_dfs(summary, experiment_params)

    def _get_primitive_objs_dict(self, summary) -> dict:
        """
        Collects all tagged primitive type objects in provided summary df.

        Parameters
        ----------
        summary: summary DataFrame of tagged variables returned from Tagr.summary()
        experiment_params: dict of experiment namespace variables

        Returns
        -------
        dict of key: obj name, val: obj

        """
        logger.info("collecting primitive objects for json file")
        primitive_objs_dict = {}
        summary_primitive_objs = summary[summary["dtype"] == "primitive"]
        for i, row in summary_primitive_objs.iterrows():
            primitive_objs_dict[row["obj_name"]] = row["val"]

        return primitive_objs_dict

    def _flush_metadata_json(self, summary, experiment_params):
        """
        Collects names and values of tagged primitive objects and metadata dataframes.
        Pushes to metadata provider as json
        """
        primitive_objs_dict = self._get_primitive_objs_dict(summary)

        df_metadata = {
            "types": self.col_types_dict,
            "stats": self.col_stats_dict,
            "primitive_objs": primitive_objs_dict,
        }

        logger.info("flushing metadata json to " + self.storage_provider.name)
        self.storage_provider.dump_json(
            df_metadata,
            experiment_params["proj"],
            experiment_params["experiment"],
            experiment_params["tag"],
        )

    def _flush_dfs(self, summary, experiment_params):
        """
        Collects summary statistics for all tagged dataframes in provided summary df.
        Saves statistics to `col_types_dict` and `col_types_dict` class attributes.
        Pushes dataframes to metadata provider as csv
        """
        summary_dfs = summary[summary["dtype"] == "dataframe"]

        logger.info("collecting dataframe types and summary stats for json file")
        for i, row in summary_dfs.iterrows():
            df = row["val"]
            df_name = row["obj_name"]
            self.col_types_dict[df_name] = dict(
                zip(df.columns, df.dtypes.map(lambda x: x.name))
            )
            self.col_stats_dict[df_name] = df.describe().to_dict()

            logger.info("flushing dataframes as csv to " + self.storage_provider.name)
            self.storage_provider.dump_csv(
                df,
                experiment_params["proj"],
                experiment_params["experiment"],
                experiment_params["tag"],
                df_name,
            )

    def _flush_non_dfs(self, summary, experiment_params):
        """
        Pushes all non dataframe, int, float and str objects to metadata provider as pickle
        """
        summary_objects = summary[summary["dtype"] != "dataframe"]
        for i, row in summary_objects.iterrows():
            obj = row["val"]
            obj_name = row["obj_name"]
            logger.info("flushing objects as pickle to " + self.storage_provider.name)
            self.storage_provider.dump_pickle(
                obj,
                experiment_params["proj"],
                experiment_params["experiment"],
                experiment_params["tag"],
                obj_name,
            )

    def list(self, proj, experiment, tag=None, dump="local"):
        """
        fetches previously flushed experiments
        Parameters
        ----------
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message (optional)
        dump: destination for experiment data to be fetched from ('aws', 'local')
            - for dump, asssume local by default if destination not provided
        """
        if dump == "aws":
            self.storage_provider = Aws()
        elif dump == "local":
            self.storage_provider = Local()

        return self.storage_provider.__list(proj, experiment, tag)

    def fetch(self, proj, experiment, tag, filename, dump="local"):
        if dump == "aws":
            self.storage_provider = Aws()
        elif dump == "local":
            self.storage_provider = Local()

        return self.storage_provider.__fetch(proj, experiment, tag, filename)



"""
Azure Blob class implementation
"""

import os
import json
import pickle

from azure.storage.blob import BlobServiceClient

from io import StringIO
from tagr.utils import NpEncoder

AZURE_STORAGE_KEY = os.environ["AZURE_STORAGE_KEY"]
AZURE_STORAGE_ACCOUNT = os.environ["AZURE_STORAGE_ACCOUNT"]


class Azure:
    def __init__(self):
        self.client = BlobServiceClient(
            account_url="https://{}.blob.core.windows.net".format(
                AZURE_STORAGE_ACCOUNT
            ),
            credential=AZURE_STORAGE_KEY,
        )
        self.name = "Azure"

    def dump_csv(self, df, proj, experiment, tag, filename):
        """
        turns dataframe into csv and saves it to Azure Storage Account directory

        Parameters
        ----------
        df: dataframe object
        proj: project name on metadata provider
        exp: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        csv_buffer = StringIO()
        df.to_csv(csv_buffer)
        container = self.client.get_container_client(proj)
        container.upload_blob(
            name="{}/{}/{}.csv".format(experiment, tag, filename),
            data=csv_buffer.getvalue(),
        )

    def dump_json(self, df_metadata, proj, experiment, tag):
        """
        turns dictionary into json and saves it to Azure Storage Account directory

        Parameters
        ----------
        df_metadata: json object containing experiment metadata
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        """
        container = self.client.get_container_client(proj)
        container.upload_blob(
            name="{}/{}/df_summary.json".format(experiment, tag),
            data=bytes(json.dumps(df_metadata, cls=NpEncoder).encode("UTF-8")),
        )

    def dump_pickle(self, model, proj, experiment, tag, filename):
        """
        serializes object to s3 directory

        Parameters
        ----------
        pickle_object: model that has been serialized into a pickle object
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        pickle_object = pickle.dumps(model)
        container = self.client.get_container_client(proj)
        container.upload_blob(
            name="{}/{}/{}.pkl".format(experiment, tag, filename), data=pickle_object
        )



"""
Aws S3 class implementation
"""

import json
import boto3
import pickle

from io import StringIO
from tagr.storage.aws_helper import AwsHelper
from tagr.utils import NpEncoder


class Aws:
    def __init__(self):
        self.S3 = boto3.resource("s3")
        self.name = "AWS"

    def dump_csv(self, df, proj, experiment, tag, filename):
        """
        turns dataframe into csv and saves it to s3 directory

        Parameters
        ----------
        df: dataframe object
        proj: project name on metadata provider
        exp: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        csv_buffer = StringIO()
        df.to_csv(csv_buffer)
        self.S3.Object(
            bucket_name=proj, key="{}/{}/{}.csv".format(experiment, tag, filename)
        ).put(Body=csv_buffer.getvalue())

    def dump_json(self, df_metadata, proj, experiment, tag):
        """
        turns dictionary into json and saves it to s3 directory

        Parameters
        ----------
        df_metadata: json object containing experiment metadata
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        """
        self.S3.Object(
            bucket_name=proj, key="{}/{}/df_summary.json".format(experiment, tag)
        ).put(
            Body=(bytes(json.dumps(df_metadata, cls=NpEncoder).encode("UTF-8"))),
            ContentType="application/json",
        )

    def dump_pickle(self, model, proj, experiment, tag, filename):
        """
        serializes oject to s3 directory

        Parameters
        ----------
        pickle_object: model that has been serialized into a pickle object
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        pickle_object = pickle.dumps(model)
        self.S3.Object(
            bucket_name=proj, key="{}/{}/{}.pkl".format(experiment, tag, filename)
        ).put(Body=pickle_object)

    def _Tagr__list(self, proj, experiment, tag):
        """
        gets list of files/folders located at {proj}/{experiment}/{tag}
        Parameters
        __________
        proj: project name (s3 bucket name)
        experiment: experiment name
        tag: custom commit message (optional)
        """
        aws_helper = AwsHelper()
        object_path = experiment
        if tag:
            object_path += "/" + tag

        return aws_helper.get_list_of_tables(proj, object_path)

    def _Tagr__fetch(self, proj, experiment, tag, filename):
        """
        fetches object located at {proj}/{experiment}/{tag}/{filename}
        Parameters
        __________
        proj: project name (s3 bucket name)
        experiment: experiment name
        tag: custom commit message
        filename: name of file to be fetched (either a .csv or .pkl file)
        """
        aws_helper = AwsHelper()
        object_path = experiment + "/" + tag + "/" + filename
        return aws_helper.get_object(proj, object_path)



OBJECTS = {
    "X_train": "dataframe",
    "X_test": "dataframe",
    "y_train": "dataframe",
    "y_test": "dataframe",
    "model": "model",
    "trained_model": "model",
    "prediction": "dataframe",
    "y_pred": "dataframe",
}

RECOGNIZED_DTYPES = ["dataframe", "primitive", "other"]



"""
GCP Cloud Storage class implementation
"""

import json
import pickle

from google.cloud import storage
from io import StringIO
from tagr.utils import NpEncoder


class Gcp:
    def __init__(self):
        self.client = storage.Client()
        self.name = "GCP"

    def dump_csv(self, df, proj, experiment, tag, filename):
        """
        turns dataframe into csv and saves it to GCS directory

        Parameters
        ----------
        df: dataframe object
        proj: project name on metadata provider
        exp: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        csv_buffer = StringIO()
        df.to_csv(csv_buffer)
        bucket = self.client.get_bucket(proj)
        blob = bucket.blob("{}/{}/{}.csv".format(experiment, tag, filename))
        blob.upload_from_string(csv_buffer.getvalue())

    def dump_json(self, df_metadata, proj, experiment, tag):
        """
        turns dictionary into json and saves it to GCS directory

        Parameters
        ----------
        df_metadata: json object containing experiment metadata
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        """
        bucket = self.client.get_bucket(proj)
        blob = bucket.blob("{}/{}/df_summary.json".format(experiment, tag))
        blob.upload_from_string(
            bytes(json.dumps(df_metadata, cls=NpEncoder).encode("UTF-8"))
        )

    def dump_pickle(self, model, proj, experiment, tag, filename):
        """
        serializes object to s3 directory

        Parameters
        ----------
        pickle_object: model that has been serialized into a pickle object
        proj: project name on metadata provider
        experiment: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        pickle_object = pickle.dumps(model)
        bucket = self.client.get_bucket(proj)
        blob = bucket.blob("{}/{}/{}.pkl".format(experiment, tag, filename))
        blob.upload_from_string(pickle_object)



import json
import pickle
import os
import logging
import pandas as pd

logger = logging.getLogger("saving experiment to local storage")


class Local:
    name = "Local"

    def dump_csv(self, df, proj, experiment, tag, filename):
        """
        turns dataframe into csv and saves it to local storage

        Parameters
        ----------
        df: dataframe object
        proj: project name
        experiment: experiment name
        tag: custom commit message
        filename: filename to be saved locally
        """
        df.to_csv(
            "{}/{}/{}/{}.csv".format(proj, experiment, tag, filename), index=False
        )

    def dump_json(self, df_metadata, proj, experiment, tag):
        """
        turns dataframe into csv and saves it to local storage

        Parameters
        ----------
        df_metadata: json object containing experiment metadata
        proj: project name
        experiment: experiment name
        tag: custom commit message
        """
        with open(
            "{}/{}/{}/df_summary.json".format(proj, experiment, tag), "w"
        ) as outfile:
            json.dump(df_metadata, outfile, default=str)

    def dump_pickle(self, model, proj, experiment, tag, filename):
        """
        turns dataframe into csv and saves it to local storage

        Parameters
        ----------
        pickle_object: model that has been serialized into a pickle object
        proj: project name
        experiment: experiment name
        tag: custom commit message
        filename: filename to be exported
        """
        pickle.dump(
            model, open("{}/{}/{}/{}.pkl".format(proj, experiment, tag, filename), "wb")
        )

    def _Tagr__list(self, proj, experiment, tag):
        """
        gets list of files/folders located at {proj}/{experiment}/{tag}

        Parameters
        __________
        proj: project name
        experiment: experiment name
        tag: custom commit message (optional)
        """
        path = proj + "/" + experiment
        if tag:
            path += "/" + tag

        folders = os.listdir(path)
        return folders

    def _Tagr__fetch(self, proj, experiment, tag, filename):
        path = proj + "/" + experiment + "/" + tag + "/" + filename
        if path.endswith(".csv"):
            obj = pd.read_csv(path)
        else:
            obj = pickle.load(open(path, "rb"))
        return obj

    def build_path(self, proj, experiment, tag):
        """
        sets up a folder directory within local storage to push metadata to

        Parameters
        ----------
        proj: project name
        experiment: experiment name
        tag: custom commit message
        """
        try:
            os.makedirs("{}/{}/{}".format(proj, experiment, tag))
        except OSError:
            logger.info(
                "The directory %s already exists. If using the tag argument, please provide a new unique identifier."
                % "{}/{}/{}".format(proj, experiment, tag)
            )



"""
code heavy helper functions for Aws.py
"""

import boto3
import pandas as pd
import pickle


class AwsHelper:
    def get_matching_s3_objects(self, bucket, object_path="", max_keys_per_request=1):
        """
        List objects in an S3 bucket.

        Parameters
        ----------
        bucket: Name of the S3 bucket.
        object_path: Path that leads to all tables
        max_keys_per_request: number of objects to list down
        """

        s3 = boto3.client("s3")
        kwargs = {"Bucket": bucket}

        # If the prefix is a single string (not a tuple of strings),
        # the filtering can be done directly in the S3 API.
        if isinstance(object_path, str):
            kwargs["Prefix"] = object_path
        else:
            kwargs["Prefix"] = str(object_path)

        kwargs["MaxKeys"] = max_keys_per_request

        while True:

            # The S3 API response is a large dict of metadata.
            # 'Contents' contains information about the listed objects.
            resp = s3.list_objects_v2(**kwargs)

            try:
                contents = resp["Contents"]
            except KeyError:
                return

            for obj in contents:
                key = obj["Key"]
                if key.startswith(object_path):
                    yield obj

            # The S3 API is paginated, returning up to 1000 keys at a time.
            # Pass the continuation token into the next response, until the
            # final page pages is reached (when this field is missing).
            try:
                kwargs["ContinuationToken"] = resp["NextContinuationToken"]
            except KeyError:
                break

    def get_matching_s3_keys(self, bucket, object_path="", max_keys_per_request=1):
        """
        Generate the keys in an S3 bucket.

        Parameters
        ----------
        bucket: Name of the S3 bucket.
        object_path: Path that leads to all tables
        max_keys_per_request: number of objects to list down
        """
        for obj in self.get_matching_s3_objects(
            bucket=bucket,
            object_path=object_path,
            max_keys_per_request=max_keys_per_request,
        ):
            yield obj["Key"]

    def get_list_of_tables(self, bucket, object_path):
        """
        Method to list down the tables under given snapshot path prefixed as:
        bucket/object_path/{filename}

        Parameters
        ----------
        bucket: name of s3 bucket
        object_path: Path that leads to all tables
        """
        s3_objects = []

        if object_path:
            for full_object_path in self.get_matching_s3_keys(
                bucket=bucket, object_path=object_path, max_keys_per_request=1
            ):
                s3_objects.append(bucket + "/" + full_object_path)

        return s3_objects

    def get_object(self, bucket, object_path):
        """
        Helper method to fetch .csv or .pkl object from s3

        Parameters
        ----------
        bucket: name of s3 bucket (project)
        object_path: Path that leads to the object being fetched
        """
        s3 = boto3.client("s3")
        response = s3.get_object(Bucket=bucket, Key=object_path)

        if object_path.endswith(".csv"):
            obj = pd.read_csv(response["Body"], index_col=0)
        else:
            serialized_model = response["Body"].read()
            obj = pickle.loads(serialized_model)
        return obj



from setuptools import setup, find_packages

setup(
    name="tagr",
    packages=find_packages(exclude=("test", "test.*")),
    version="0.0.2",
    license="MIT",
    long_description="Cloud Agnostic Data Science Productivity Tool",
    author="Sunrise Long",
    author_email="sunrise.long@yahoo.ca",
    url="https://github.com/tagr-dev/tagr",
    install_requires=[
        "numpy",
        "pandas",
        "boto3",
        "azure-storage-blob",
        "google-cloud-storage",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
